﻿// Crystal Dynamics Hash Generator
// Copyright 2018-2019 Ben Lincoln
// https://www.thelostworlds.net/
//

// This file is part of Crystal Dynamics Hash Generator.

// Crystal Dynamics Hash Generator is free software: you can redistribute it and/or modify
// it under the terms of version 3 of the GNU General Public License as published by
// the Free Software Foundation.

// Crystal Dynamics Hash Generator is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with Crystal Dynamics Hash Generator (in the file LICENSE.txt).  
// If not, see <http://www.gnu.org/licenses/>.

using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Text;

using CDHashes = BenLincoln.TheLostWorlds.CDHashes;
using BLD = BenLincoln.Data;

namespace CDHashGen
{
    class Program
    {
        protected const string ProgramName = "Crystal Dynamics Hash Generator";
        protected const string Copyright = "Copyright 2019 Ben Lincoln and Andrew Fradley";
        protected const string InfoURL = "https://www.thelostworlds.net/Software/";
        protected const int ReadBufferLineCount = 10000;

        static void Main(string[] args)
        {
            if ((args.Length < 3))
            {
                DisplayHelpMessage();
                Environment.Exit(1);
            }

            CDHashes.HashType hashType = CDHashes.HashType.Undefined;
            string inputPath = "";
            string matchListPath = "";
            string outputPath = "";
            string separator = "\t";
            bool hashRawText = false;
            bool bigEndian = false;
            bool hexOutput = true;
            bool base64Output = false;
            bool appendToExisting = false;
            bool forceOverwrite = false;

            ArrayList arguments = new ArrayList();
            for (int i = 0; i < args.Length; i++)
            {
                // if the argument is one with a second component, catch that now
                string nextArg = "";
                switch (args[i])
                {
                    case "--hashtype":
                    case "-t":
                        try
                        {
                            nextArg = args[i + 1];
                            nextArg = nextArg.ToLower().Trim();
                            switch (nextArg)
                            {
                                case "bloodomen": 
                                case "bo":
                                    hashType = CDHashes.HashType.BloodOmen;
                                    break;
                                case "proto1":
                                    hashType = CDHashes.HashType.SoulReaverProto1;
                                    break;
                                case "soulreaveralpha":
                                case "sra":
                                    hashType = CDHashes.HashType.SoulReaverAlpha;
                                    break;
                                case "soulreaver":
                                case "sr":
                                    hashType = CDHashes.HashType.SoulReaver;
                                    break;
                                case "soulreaverdreamcast":
                                case "srdc":
                                    hashType = CDHashes.HashType.SoulReaverDreamcast;
                                    break;
                                case "soulreaver2airforge":
                                case "sr2af":
                                    hashType = CDHashes.HashType.SoulReaver2AirForge;
                                    break;
                                case "soulreaver2":
                                case "sr2":
                                    hashType = CDHashes.HashType.SoulReaver2;
                                    break;
                                case "defiance":
                                    hashType = CDHashes.HashType.Defiance;
                                    break;
                            }
                            i++;
                        }
                        catch (Exception ex)
                        {
                            ExitWithError(string.Format("Unable to parse '{0}' as a recognized hash type.", nextArg), 3);
                        }
                        break;
                    case "--input":
                    case "-i":
                        try
                        {
                            nextArg = args[i + 1];
                            inputPath = nextArg;
                            i++;
                        }
                        catch (Exception ex)
                        {
                            ExitWithError(string.Format("Unable to parse '{0}' as an input file path.", nextArg), 3);
                        }
                        break;
                    case "--match":
                    case "-m":
                        try
                        {
                            nextArg = args[i + 1];
                            matchListPath = nextArg;
                            i++;
                        }
                        catch (Exception ex)
                        {
                            ExitWithError(string.Format("Unable to parse '{0}' as a match list file path.", nextArg), 3);
                        }
                        break;
                    case "--output":
                    case "-o":
                        try
                        {
                            nextArg = args[i + 1];
                            outputPath = nextArg;
                            i++;
                        }
                        catch (Exception ex)
                        {
                            ExitWithError(string.Format("Unable to parse '{0}' as an output file path.", nextArg), 3);
                        }
                        break;
                    case "--separator":
                    case "-s":
                        try
                        {
                            nextArg = args[i + 1];
                            separator = nextArg;
                            i++;
                        }
                        catch (Exception ex)
                        {
                            ExitWithError(string.Format("Unable to parse '{0}' as a field separator.", nextArg), 3);
                        }
                        break;
                    default:
                        arguments.Add(args[i]);
                        break;
                }
            }

            if (hashType == CDHashes.HashType.Undefined)
            {
                DisplayHelpMessage();
                ExitWithError("A hash type was not specified", 1);
            }

            ArrayList positionalArgs = new ArrayList();

            foreach (string arg in arguments)
            {
                if ((arg == "-?") || (arg == "-h") || (arg == "--help"))
                {
                    DisplayHelpMessage();
                    Environment.Exit(0);
                }
                bool handled = false;
                if (arg == "--hash-raw-text")
                {
                    hashRawText = true;
                    handled = true;
                }
                if (arg == "--big-endian")
                {
                    bigEndian = true;
                    handled = true;
                }
                if (arg == "--base64")
                {
                    base64Output = true;
                    handled = true;
                }
                if (arg == "--append")
                {
                    appendToExisting = true;
                    handled = true;
                }
                if (arg == "--overwrite")
                {
                    forceOverwrite = true;
                    handled = true;
                }
                // right now, all positional arguments are strings to be hashed
                if (!handled)
                {
                    positionalArgs.Add(arg);
                }
            }

            string[] inputStrings = new string[] { };
            string[] matchHashes = new string[] { };

            if (inputPath != "")
            {
                if (!File.Exists(inputPath))
                {
                    ExitWithError(string.Format("The file '{0}' could not be found.", inputPath), 4);
                }
                //inputStrings = GetLinesAsStringArray(GetFileAsString(inputPath));
            }

            if (matchListPath != "")
            {
                if (!File.Exists(matchListPath))
                {
                    ExitWithError(string.Format("The file '{0}' could not be found.", matchListPath), 4);
                }
                matchHashes = GetLinesAsStringArray(GetFileAsString(matchListPath));
                for (int i = 0; i < matchHashes.Length; i++)
                {
                    matchHashes[i] = matchHashes[i].Trim().ToUpper();
                }
            }

            if (outputPath != "")
            {
                if (!appendToExisting)
                {
                    if (File.Exists(outputPath))
                    {
                        if (!forceOverwrite)
                        {
                            ExitWithError(string.Format("The output file '{0}' already exists. Use the --append option to add to the end of the existing file, or --overwrite to replace the file.", outputPath), 5);
                        }
                        else
                        {
                            try
                            {
                                FileStream createStream = new FileStream(outputPath, FileMode.Create, FileAccess.Write);
                                createStream.Close();
                            }
                            catch (Exception ex)
                            {
                                ExitWithError(string.Format("Could not write to the output file '{0}': {1}", outputPath, ex.Message), 10);
                            }
                        }
                    }
                }
            }

            StringBuilder outBuilder = new StringBuilder();

            CDHashes.CDHash hash = new CDHashes.CDHash();

            switch (hashType)
            {
                case CDHashes.HashType.BloodOmen:
                    hash = new CDHashes.BloodOmenHash();
                    break;
                case CDHashes.HashType.SoulReaverProto1:
                    hash = new CDHashes.SoulReaverProto1DemoHash();
                    break;
                case CDHashes.HashType.SoulReaverAlpha:
                    hash = new CDHashes.SoulReaverAlphaHash();
                    break;
                case CDHashes.HashType.SoulReaver:
                    hash = new CDHashes.SoulReaverHash();
                    break;
                case CDHashes.HashType.SoulReaverDreamcast:
                    hash = new CDHashes.SoulReaverDreamcastHash();
                    break;
                case CDHashes.HashType.SoulReaver2AirForge:
                    hash = new CDHashes.SoulReaver2AirForgeHash();
                    break;
                case CDHashes.HashType.SoulReaver2:
                    hash = new CDHashes.SoulReaver2Hash();
                    break;
                case CDHashes.HashType.Defiance:
                    hash = new CDHashes.DefianceHash();
                    break;
            }

            foreach (string inString in positionalArgs)
            {
                GetHash(hash, inString, matchHashes, outputPath, separator, hexOutput, base64Output, bigEndian);
            }

            if (inputPath != "")
            {
                FileStream inStream = new FileStream(inputPath, FileMode.Open, FileAccess.Read);
                StreamReader inReader = new StreamReader(inStream);

                bool done = false;

                while (!done)
                {
                    inputStrings = new string[ReadBufferLineCount];
                    int numLines = 0;
                    try
                    {
                        for (int i = 0; i < ReadBufferLineCount; i++)
                        {
                            if (inReader.EndOfStream)
                            {
                                break;
                            }
                            inputStrings[i] = inReader.ReadLine().TrimEnd();
                            numLines++;
                        }
                    }
                    catch (Exception ex)
                    {
                        done = true;
                        ExitWithError(string.Format("An exception occurred while reading from the source file: {0}", ex.Message), 17);
                    }
                    for (int i = 0; i < numLines; i++)
                    {
                        {
                            GetHash(hash, inputStrings[i], matchHashes, outputPath, separator, hexOutput, base64Output, bigEndian);
                        }
                    }
                    if (inReader.EndOfStream)
                    {
                        done = true;
                    }
                }

                inReader.Close();
                inStream.Close();
            }

            Console.WriteLine("");
        }

        protected static void GetHash(CDHashes.CDHash h, string inputString, string[] matchList, string outputPath, string separator, bool hexOutput, bool base64Output, bool bigEndian)
        {
            StringBuilder builder = new StringBuilder();

            byte[] hashValue = h.GetHashBytes(inputString, bigEndian);
            string outStringHex = BLD.HexConverter.ByteArrayToHexString(hashValue);
            bool writeOutput = false;
            if (matchList.Length > 0)
            {
                if (matchList.Contains(outStringHex))
                {
                    writeOutput = true;
                }
            }
            else
            {
                writeOutput = true;
            }
            if (writeOutput)
            {
                if (hexOutput)
                {
                    builder.Append(outStringHex);
                    builder.Append(separator);
                }
                builder.Append(inputString);
                builder.Append(separator);
               if (base64Output)
                {
                    builder.Append(Convert.ToBase64String(hashValue));
                    builder.Append(separator);
                }
            }
            string outputString = builder.ToString();
            if (outputString != "")
            {
                if (outputPath != "")
                {
                    try
                    {
                        WriteToOutput(outputPath, outputString, FileMode.Append);
                    }
                    catch (Exception ex)
                    {
                        Console.WriteLine(string.Format("Error writing to output file '{0}': {1}", outputPath, ex.Message));
                    }
                }
                Console.WriteLine(outputString);
            }
        }

        static void DisplayHelpMessage()
        {
            DisplayBanner();

            Assembly asm = Assembly.GetExecutingAssembly();
            Stream inStream = asm.GetManifestResourceStream("CDHashGen.HelpMessage.txt");
            StreamReader inReader = new StreamReader(inStream);
            Console.Write(inReader.ReadToEnd());
            Console.WriteLine();
            inReader.Close();
            inStream.Close();
        }

        static void DisplayBanner()
        {
            Assembly currentAssembly = Assembly.GetExecutingAssembly();

            AssemblyName assemblyName = currentAssembly.GetName();

            int versionMajor = assemblyName.Version.Major;
            int versionMinor = assemblyName.Version.Minor;

            Console.WriteLine(string.Format("{0} {1}.{2}", ProgramName, versionMajor, versionMinor));
            Console.WriteLine(Copyright);
            Console.WriteLine(InfoURL);
            Console.WriteLine(Environment.NewLine);
        }

        protected static void ExitWithError(string message, int exitCode)
        {
            DisplayBanner();
            Console.WriteLine(message);
            System.Environment.ExitCode = exitCode;
            Environment.Exit(exitCode);
        }

        private static void WriteToOutput(string outputPath, string text, FileMode mode)
        {
            FileStream oStream;
            StreamWriter oWriter;

            oStream = new FileStream(outputPath, mode, FileAccess.Write);
            oWriter = new StreamWriter(oStream);

            oWriter.WriteLine(text);

            oWriter.Close();
            oStream.Close();
        }

        private static string GetFileAsString(string inputPath)
        {
            string result = "";
            FileStream inStream;
            StreamReader inReader;

            inStream = new FileStream(inputPath, FileMode.Open, FileAccess.Read);
            inReader = new StreamReader(inStream);

            result = inReader.ReadToEnd();

            inReader.Close();
            inStream.Close();
            return result;
        }

        private static string[] GetLinesAsStringArray(string inputString)
        {
            // for easier cross-platform use, this handles Unix and DOS-style newlines the same way
            string inString = inputString.Replace("\r\n", "\n");

            string[] result = inString.Split(new string[] {"\n"} , StringSplitOptions.RemoveEmptyEntries);

            return result;
        }
    }
}
